print("import: 'pydl'")
import pydl

print("import: 'pydl.goddard'")
import pydl.goddard

print("import: 'pydl.photoop'")
import pydl.photoop

print("import: 'pydl.pydlspec2d'")
import pydl.pydlspec2d

print("import: 'pydl.pydlutils'")
import pydl.pydlutils

