# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import os
import numpy as np
from astropy.tests.helper import raises
from ..window import sdss_score, window_read, window_score


class TestWindow(object):
    """Test the functions in pydl.photoop.window.
    """

    def setup(self):
        pass

    def teardown(self):
        pass

    def test_sdss_score(self):
        pass

    def test_window_read(self):
        pass

    def test_window_score(self):
        pass
